/*   This is part of ALE4Net
 *   Application Level Environment for Networking
 *   
 *   support subroutines: does they really need?
 *
 *   Copyright 2004 Renzo Davoli University of Bologna - Italy
 *   
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License along
 *   with this program; if not, write to the Free Software Foundation, Inc.,
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */   
/*
 * Neelkanth Natu (neelnatu@yahoo.com) is the author of this file.
 * As long as you keep this blurb in the file, feel free to do whatever.
 */ 
#include <sys/types.h>
#include <stdio.h>
#include <assert.h>
#include <string.h>
#include <stdlib.h>
#include <sys/time.h>
#include <sys/signal.h>
#include <sys/syscall.h>
#include <errno.h>

#include "alpine_support.h"

int 
alpine_bcmp(const void *s1, const void *s2, size_t n)
{
	return (bcmp(s1, s2, n));
}

void alpine_bcopy(const void *src, void *dest, size_t n)
{
	bcopy(src, dest, n);
}

void 
alpine_bzero(void *s, size_t n)
{
	bzero(s, n);
}

/* XXX: Avoid an infinite recursion between alpine_malloc() and malloc() */
extern void *__libc_malloc(size_t size);
extern void *__libc_calloc(size_t nmemb, size_t size);
extern void *__libc_realloc(void *ptr, size_t size);
extern void __libc_free(void *ptr);

void *
alpine_malloc(size_t size)
{
	return (__libc_malloc(size));
}

void *
alpine_calloc(size_t nmemb, size_t size)
{
	return (__libc_calloc(nmemb, size));
}

void *
alpine_realloc(void *ptr, size_t size)
{
	return (__libc_realloc(ptr, size));
}

void 
alpine_free(void *ptr)
{
	__libc_free(ptr);
}

int 
alpine_gettimeofday(struct timeval *tv, struct timezone *tz)
{
	int rc;

	rc = gettimeofday(tv, tz);
	assert(rc == 0);
}

void *
alpine_memcpy(void *dest, const void *src, size_t n)
{
	return (memcpy(dest, src, n));
}

void *
alpine_memmove(void *dest, const void *src, size_t n)
{
	return (memmove(dest, src, n));
}

int 
alpine_putchar(int c)
{
	return (putchar(c));
}

int 
alpine_puts(const char *s)
{
	return (puts(s));
}

int 
alpine_vprintf(const char *format, va_list ap)
{
	return (vprintf(format, ap));
}

int 
alpine_vsprintf(char *str, const char *format, va_list ap)
{
	return (vsprintf(str, format, ap));
}

int 
alpine_vsnprintf(char *str, size_t size, const char *format, va_list ap)
{
	return (vsnprintf(str, size, format, ap));
}

int
alpine_sigpipe(void)
{
	return (kill(getpid(), SIGPIPE));
}
