/*   This is part of ALE4Net
 *   Application Level Environment for Networking
 *
 *   fd_mgmt: access control and fd management
 *   
 *   Copyright 2004 Renzo Davoli University of Bologna - Italy
 *   
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License along
 *   with this program; if not, write to the Free Software Foundation, Inc.,
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */   
#include <linux/limits.h>
#include <sys/select.h>
#include <unistd.h>
#include <stdlib.h>

#define FALSE 0

static unsigned char opencount[OPEN_MAX];
static fd_set *fds[OPEN_MAX];


void
fdmgmt_open(int controlfd, int fd)
{
	//printf("fdmgmt_open %d %d %d %d\n",controlfd,fd,sizeof(fd_set),sizeof(fds));
	if (fd >= 0) {
		if (opencount[controlfd]==0) {
			fds[controlfd]=(fd_set *) malloc (sizeof (fd_set));
			FD_ZERO(fds[controlfd]);
		}
		opencount[fd]++;
		opencount[controlfd]++;
		FD_SET(fd,fds[controlfd]);
	}
}

int
fdmgmt_close(int controlfd, int fd, int (*handler)(int fd))
{
	//printf("fdmgmt_close %d %d\n",controlfd,fd);
	if (fd >= 0 && opencount[fd] > 0 && opencount[controlfd] > 0 && fds[controlfd] != NULL) {
		if (FD_ISSET(fd,fds[controlfd])) {
			FD_CLR(fd,fds[controlfd]);
			if (--opencount[fd] == 0) {
				//printf("eliminate %d\n",fd);
				return handler(fd);
			}
			if (--opencount[controlfd] == 0) {
				free(fds[controlfd]);
				fds[controlfd]=NULL;
			}
		}
	}
	return handler(fd);
}

void
fdmgmt_controlclose(int controlfd,int (*handler)(int fd))
{
	register int i;
	//printf("fdmgmt_controlclose %d\n",controlfd);
	if (opencount[controlfd] > 0 && fds[controlfd] != NULL) {
		for (i=0;i<OPEN_MAX;i++) {
			if (FD_ISSET(i,fds[controlfd]))
				fdmgmt_close(controlfd,i,handler);
		}
	}
}

void
fdmgmt_controldup(int controlfd,int newcontrolfd)
{
	register int i;
	//printf("fdmgmt_controldup %d %d\n",controlfd,newcontrolfd);
	if (opencount[controlfd] > 0 && fds[controlfd] != NULL) {
		for (i=0;i<OPEN_MAX;i++) {
			if (FD_ISSET(i,fds[controlfd])) 
				fdmgmt_open(controlfd,i);
		}
	}
}

int
fdmgmt_check(int controlfd,int fd)
{
	//printf("fdmgmt_check %d %d\n",controlfd,fd);
	if (opencount[controlfd]  > 0 && fds[controlfd] != NULL) 
		return FD_ISSET(fd,fds[controlfd]);
	else
		return FALSE;
}
