/*   This is part of ALE4Net
 *   Application Level Environment for Networking
 *
 *   client_socket: redefinition of system calls
 *   
 *   Copyright 2004 Renzo Davoli University of Bologna - Italy
 *   
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License along
 *   with this program; if not, write to the Free Software Foundation, Inc.,
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */   
/*
 * Neelkanth Natu (neelnatu@yahoo.com) is the author of this file.
 * As long as you keep this blurb in the file, feel free to do whatever.
 */ 
#include <sys/param.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <errno.h>
#include <assert.h>
#include <sys/queue.h>
#include <sys/ioctl.h>
#include <net/if.h>
#include <fcntl.h>
#include <sys/time.h>
#include <sys/select.h>
#include <sys/uio.h>
#include <sys/un.h>
#include <unistd.h>

#include <stdarg.h>

#include "fd_map.h"
#include "req_resp.h"

#ifndef TIMEVAL_TO_TIMESPEC
#define TIMEVAL_TO_TIMESPEC(tv, ts) {					\
	(ts)->tv_sec = (tv)->tv_sec;					\
	(ts)->tv_nsec = (tv)->tv_usec * 1000;				\
}
#endif	/* ifndef TIMEVAL_TO_TIMESPEC */

#ifndef TIMESPEC_TO_TIMEVAL
#define TIMESPEC_TO_TIMEVAL(tv, ts) {					\
	(tv)->tv_sec = (ts)->tv_sec;					\
	(tv)->tv_usec = (ts)->tv_nsec / 1000;				\
}
#endif	/* ifndef TIMESPEC_TO_TIMEVAL */

#define INIT_LIB() 							\
	do { 								\
		if (!lib_inited) 					\
			do_init(); 					\
		if (lib_init_failed) {					\
			errno = EAGAIN;					\
			return (-1);					\
		}							\
	} while (0)

/* XXX: close() all the native_fds in fd_map? */
#define CLEANUP_LIB()							\
	do {								\
		if (connect_fd >= 0)					\
			close(connect_fd);				\
		connect_fd = -1;					\
		connect_pid = -1;					\
		lib_inited = 0;						\
		lib_init_failed = 0;					\
	} while (0)

#define CLEANUP_LIB_AND_RETURN()					\
	do {								\
		CLEANUP_LIB();						\
		errno = EAGAIN;						\
		return (-1);						\
	} while (0)

/* 
 * XXX: We cannot include freebsd_errno.h because it does not like
 * _POSIX_SOURCE defined. Unfortunately one of our include files defines
 * it.
 */
//extern int map_errno(int freebsd_errno);
#define map_errno(x) (x)
static int lib_inited;
static int lib_init_failed;

static int connect_fd = -1;
static int connect_pid = -1;

static const char *default_connect_addr = "127.0.0.1";
static const int default_connect_port = 8475;
static const char *default_connect_path = "/tmp/a4ns";
static const char *connect_addr = NULL;
static int connect_port = -1;

static int do_init(void);
//static int do_connect(const char *addr, int port, int32_t *ret_pid);
static int do_connect(const char *addr, int32_t *ret_pid);
static int get_native_fd(void);

static int getname(int native_fd, struct sockaddr *name, socklen_t *ret_namelen,
		   enum request_type req, enum response_type resp);

static int recvfrom_internal(int native_fd, void *buf, size_t len, int flags,
		  	     struct sockaddr *from, socklen_t *fromlen,
		  	     enum request_type req, enum response_type resp);

static int sendto_internal(int s, const void *buf, size_t len, int flags,
			   const struct sockaddr *to, int tolen,
			   enum request_type req, enum response_type resp);

static int freebsd_dup(int old_fd);

/* 
 * XXX: The following two functions populate_alpine_fdset() and
 * populate_native_fdset() are similar in names but their return values
 * are completely different. Beware.
 */

/*
 * Initialize the alpine_fdsets by mapping fds from native_fdsets.
 * Return the maximum alpine_fd encountered while mapping the fdsets or
 * -1 if none of the fds was an alpine_fd.
 */
static int populate_alpine_fdset(int nfd, 
	fd_set *native_rdset, fd_set *native_wrset, fd_set *native_exset,
	fd_set *alpine_rdset, fd_set *alpine_wrset, fd_set *alpine_exset);

/*
 * Populate the native_fdsets from alpine_fdsets.
 * Returns the total number of fds mapped into all the native_fdsets.
 */
static int populate_native_fdset(int alpine_nfd, 
				const fd_set *alpine_rdset, 
				const fd_set *alpine_wrset, 
				const fd_set *alpine_exset,
				fd_set *native_rdset, 
				fd_set *native_wrset, 
				fd_set *native_exset);
int
do_init(void)
{
	assert(lib_inited == 0);
	assert(lib_init_failed == 0);

	lib_inited = 1;
	
	fd_map_init();

/*	connect_addr = getenv("ALPINE_SERVER_LISTEN_ADDR") ?
		       (const char *)getenv("ALPINE_SERVER_LISTEN_ADDR") :
		       default_connect_addr;
	connect_port = getenv("ALPINE_SERVER_LISTEN_PORT") ?
		       atoi(getenv("ALPINE_SERVER_LISTEN_PORT")) :
		       default_connect_port;

	connect_fd = do_connect(connect_addr, connect_port, &connect_pid);*/
	connect_addr = getenv("ALPINE_SERVER_PATH") ?
		       (const char *)getenv("ALPINE_SERVER_PATH") :
		       default_connect_path;
	connect_fd = do_connect(connect_addr, &connect_pid);

end:
	if (connect_fd >= 0) {
		return (0);
	} else {
		lib_init_failed = 1;
		return (-1);
	}
}

/*int
do_connect(const char *addr, int port, int32_t *ret_pid)*/
int
do_connect(const char *addr, int32_t *ret_pid)
{
	//struct sockaddr_in sin;
	struct sockaddr_un sin;
	int n, fd;

	//fd = native_socket(AF_INET, SOCK_STREAM, 0);
	fd = native_socket(AF_UNIX, SOCK_STREAM, 0);
	if (fd == -1) {
		printf("Error opening socket: %s\n", strerror(errno));
		return (-1);
	}

	bzero(&sin, sizeof(sin));
	/*sin.sin_family = AF_INET;
	sin.sin_addr.s_addr = inet_addr(addr);
	sin.sin_port = htons(port);*/
	sin.sun_family = AF_UNIX;
	strcpy(sin.sun_path,addr);
	n = native_connect(fd, (struct sockaddr *)&sin, sizeof(sin));
	if (n != 0) {
		printf("connect() error : %s\n", strerror(errno));
		native_close(fd);
		return (-2);
	}

	n = read_n_bytes(fd, (char *)ret_pid, sizeof(*ret_pid));
	if (n != sizeof(*ret_pid)) {
		printf("Error reading our pid from the server: %s\n",
			strerror(errno));
		native_close(fd);
		return (-3);
	}

	assert(*ret_pid >= 0);

	return (fd);
}

/*
 * XXX: Check if we have a mapping for this fd. I was not expecting
 * that a native_fd would be close()d from underneath us (since we
 * overload the close() function). Unfortunately though it happens.
 * For example: res_nclose() in glibc calls __close() that is
 * mapped to the glibc internal function. glibc uses something
 * called a libc_hidden_symbol() in //glibc/include/libc-symbols.h
 * that can be used to prevent overriding of certain functions and
 * global variables.
 */
int
get_native_fd(void)
{
	int native_fd, alpine_fd;
	int ntries = 0;

	while (ntries++ < 10) {
#if 0
		native_fd = open("/dev/null", 0);
#else
		/* 
		 * This is so that fstat() on the native_fd will return
		 * the right file type. Some programs depends on the fd
		 * being S_ISSOCK(); e.g. vsftpd.
		 */
		native_fd = native_socket(AF_INET, SOCK_DGRAM, 0);
#endif
		if (native_fd == -1) {
			return (-1);
		}

		alpine_fd = fd_map_lookup_native_fd(native_fd);
		if (alpine_fd == -1) {
			return (native_fd);	/* common case */
		}

		/* Close native_fd (this will close alpine_fd implicitly) */
		(void) close(native_fd);
	}

	printf("Unable to close a lost mapping between native and alpine fd\n");
	return (-1);
}

int
socket(int domain, int type, int protocol)
{
	int native_fd, nread, nwritten;
	struct socket_req socket_req;
	struct socket_resp socket_resp;

	INIT_LIB();

	/* 
	 * If the socket domain (af_*) is something that we don't
	 * understand, then use native_socket()
	 */
	//if (linux_to_bsd_domain(domain) == -1) {
	if (domain != PF_INET && domain != PF_INET6) {
		return (native_socket(domain, type, protocol));
	}

	native_fd = get_native_fd();
	if (native_fd == -1) {
		return (-1);
	}

	socket_req.req_type = REQ_SOCKET;
	//socket_req.domain = linux_to_bsd_domain(domain);
	socket_req.domain = domain;
	socket_req.type = type;
	socket_req.protocol = protocol;

	/* Write out the request */
	nwritten = write_n_bytes(connect_fd, 
				 (char *)&socket_req, sizeof(socket_req));
	if (nwritten != sizeof(socket_req)) {
		native_close(native_fd);
		CLEANUP_LIB_AND_RETURN();
	}

	/* Now wait for the response */
	nread = read_n_bytes(connect_fd, 
			     (char *)&socket_resp, sizeof(socket_resp));
	if (nread != sizeof(socket_resp)) {
		native_close(native_fd);
		CLEANUP_LIB_AND_RETURN();
	}
	assert(socket_resp.resp_type == RESP_SOCKET);
	
	/* Create a mapping between native_fd and socket_fd */
	if (socket_resp.retval >= 0) {
		fd_map_create(native_fd, socket_resp.retval);
		return (native_fd);
	} else {
		native_close(native_fd);
		errno = map_errno(socket_resp.alpine_errno);
		return (-1);
	}

}

int
close(int native_fd)
{
	struct close_req close_req;
	struct close_resp close_resp;
	int nread, nwritten;
	int alpine_fd, retval = -1;

	alpine_fd = fd_map_lookup_native_fd(native_fd);

	retval = native_close(native_fd);

	if (alpine_fd != -1) {
		fd_map_destroy(native_fd);

		/* Send the request */
		close_req.req_type = REQ_CLOSE;
		close_req.fd = alpine_fd;
		nwritten = write_n_bytes(connect_fd,
					 (char *)&close_req, 
					 sizeof(close_req));
		if (nwritten != sizeof(close_req)) {
			CLEANUP_LIB_AND_RETURN();
		}

		/* Read the response */
		nread = read_n_bytes(connect_fd,
				     (char *)&close_resp,
				     sizeof(close_resp));
		if (nread != sizeof(close_resp)) {
			CLEANUP_LIB_AND_RETURN();
		}
		assert(close_resp.resp_type == RESP_CLOSE);
		retval = close_resp.retval;
		if (retval == -1) {
			errno = map_errno(close_resp.alpine_errno);
		}
	}
	
	return (retval);
}

int
shutdown(int native_fd, int how)
{
	struct shutdown_req shutdown_req;
	struct shutdown_resp shutdown_resp;
	int nread, nwritten;
	int alpine_fd, retval = -1;

	alpine_fd = fd_map_lookup_native_fd(native_fd);

	if (alpine_fd == -1) {
		return (native_shutdown(native_fd, how));
	}

	/* Send the request */
	shutdown_req.req_type = REQ_SHUTDOWN;
	shutdown_req.fd = alpine_fd;
	shutdown_req.how = how;
	nwritten = write_n_bytes(connect_fd,
				 (const char *)&shutdown_req, 
				 sizeof(shutdown_req));
	if (nwritten != sizeof(shutdown_req)) {
		CLEANUP_LIB_AND_RETURN();
	}

	/* Read the response */
	nread = read_n_bytes(connect_fd,
			     (char *)&shutdown_resp,
			     sizeof(shutdown_resp));
	if (nread != sizeof(shutdown_resp)) {
		CLEANUP_LIB_AND_RETURN();
	}
	assert(shutdown_resp.resp_type == RESP_SHUTDOWN);
	retval = shutdown_resp.retval;
	if (retval == -1) {
		errno = map_errno(shutdown_resp.alpine_errno);
	}
	
	return (retval);
}

int
bind(int native_fd, const struct sockaddr *myaddr, socklen_t addrlen)
{
	struct bind_req bind_req;
	struct bind_resp bind_resp;
	int nread, nwritten;
	int alpine_fd, retval = -1;

	if ((myaddr == NULL) || (addrlen <= 0)) {
		errno = EINVAL;
		return (-1);
	}

	/* Is it one of ours? */
	alpine_fd = fd_map_lookup_native_fd(native_fd);
	if (alpine_fd == -1) {
		return (native_bind(native_fd, myaddr, addrlen));
	}

	bind_req.req_type = REQ_BIND;
	bind_req.fd = alpine_fd;
	bind_req.addrlen = addrlen;

	/* Write out the request */
	nwritten = write_n_bytes(connect_fd, 
				 (char *)&bind_req, sizeof(bind_req));
	if (nwritten != sizeof(bind_req)) {
		CLEANUP_LIB_AND_RETURN();
	}

	nwritten = write_n_bytes(connect_fd, (char *)myaddr, addrlen);
	if (nwritten != addrlen) {
		CLEANUP_LIB_AND_RETURN();
	}

	/* Read the response */
	nread = read_n_bytes(connect_fd,
			     (char *)&bind_resp, sizeof(bind_resp));
	if (nread != sizeof(bind_resp)) {
		CLEANUP_LIB_AND_RETURN();
	}
	assert(bind_resp.resp_type == RESP_BIND);

	/* Set errno */
	if (bind_resp.retval == -1) {
		errno = map_errno(bind_resp.alpine_errno);
	}

	return (bind_resp.retval);
}

int
connect(int native_fd, const struct sockaddr *addr, socklen_t addrlen)
{
	struct connect_req connect_req;
	struct connect_resp connect_resp;
	int nread, nwritten;
	int alpine_fd, retval = -1;

	if ((addr == NULL) || (addrlen <= 0)) {
		errno = EINVAL;
		return (-1);
	}

	/* Is it one of ours? */
	alpine_fd = fd_map_lookup_native_fd(native_fd);
	if (alpine_fd == -1) {
		return (native_connect(native_fd, addr, addrlen));
	}

	connect_req.req_type = REQ_CONNECT;
	connect_req.fd = alpine_fd;
	connect_req.addrlen = addrlen;

	/* Write out the request */
	nwritten = write_n_bytes(connect_fd, 
				 (char *)&connect_req, sizeof(connect_req));
	if (nwritten != sizeof(connect_req)) {
		CLEANUP_LIB_AND_RETURN();
	}

	nwritten = write_n_bytes(connect_fd, (char *)addr, addrlen);
	if (nwritten != addrlen) {
		CLEANUP_LIB_AND_RETURN();
	}

	/* Read the response */
	nread = read_n_bytes(connect_fd,
			     (char *)&connect_resp, sizeof(connect_resp));
	if (nread != sizeof(connect_resp)) {
		CLEANUP_LIB_AND_RETURN();
	}
	assert(connect_resp.resp_type == RESP_CONNECT);

	/* Set errno */
	if (connect_resp.retval == -1) {
		errno = map_errno(connect_resp.alpine_errno);
	}

	return (connect_resp.retval);
}

int
listen(int native_fd, int backlog)
{
	struct listen_req listen_req;
	struct listen_resp listen_resp;
	int nread, nwritten;
	int alpine_fd, retval = -1;

	/* Is it one of ours? */
	alpine_fd = fd_map_lookup_native_fd(native_fd);
	if (alpine_fd == -1) {
		return (native_listen(native_fd, backlog));
	}

	listen_req.req_type = REQ_LISTEN;
	listen_req.fd = alpine_fd;
	listen_req.backlog = backlog;

	/* Write out the request */
	nwritten = write_n_bytes(connect_fd, 
				 (char *)&listen_req, sizeof(listen_req));
	if (nwritten != sizeof(listen_req)) {
		CLEANUP_LIB_AND_RETURN();
	}

	/* Read the response */
	nread = read_n_bytes(connect_fd,
			     (char *)&listen_resp, sizeof(listen_resp));
	if (nread != sizeof(listen_resp)) {
		CLEANUP_LIB_AND_RETURN();
	}
	assert(listen_resp.resp_type == RESP_LISTEN);

	/* Set errno */
	if (listen_resp.retval == -1) {
		errno = map_errno(listen_resp.alpine_errno);
	}

	return (listen_resp.retval);
}

int
read(int native_fd, void *data, size_t count)
{
	struct read_req read_req;
	struct read_resp read_resp;
	int nread, nwritten;
	int alpine_fd, retval = -1;

	/* Is it one of ours? */
	alpine_fd = fd_map_lookup_native_fd(native_fd);
	if (alpine_fd == -1) {
		return (native_read(native_fd, data, count));
	}

	read_req.req_type = REQ_READ;
	read_req.fd = alpine_fd;
	read_req.count = count;

	/* Write out the request */
	nwritten = write_n_bytes(connect_fd, 
				 (char *)&read_req, sizeof(read_req));
	if (nwritten != sizeof(read_req)) {
		CLEANUP_LIB_AND_RETURN();
	}

	/* Read the response */
	nread = read_n_bytes(connect_fd,
			     (char *)&read_resp, sizeof(read_resp));
	if (nread != sizeof(read_resp)) {
		CLEANUP_LIB_AND_RETURN();
	}
	assert(read_resp.resp_type == RESP_READ);

	/* And the data if appropriate */
	if (read_resp.retval > 0) {
		nread = read_n_bytes(connect_fd, data, read_resp.retval);
		if (nread != read_resp.retval) {
			CLEANUP_LIB_AND_RETURN();
		}
	}

	/* Set errno */
	if (read_resp.retval == -1) {
		errno = map_errno(read_resp.alpine_errno);
	}

	return (read_resp.retval);
}

int
recv(int native_fd, void *buf, size_t len, int flags)
{
	return(recvfrom_internal(native_fd, buf, len, flags, 
				 NULL, NULL, REQ_RECV, RESP_RECV));
}

int
recvfrom(int native_fd, void *buf, size_t len, int flags,
	 struct sockaddr *from, socklen_t *fromlen)
{
	return(recvfrom_internal(native_fd, buf, len, flags, 
				 from, fromlen, REQ_RECVFROM, RESP_RECVFROM));
}

int
recvmsg(int native_fd, struct msghdr *msg, int flags)
{
	struct recvmsg_req recvmsg_req;
	struct recvmsg_resp recvmsg_resp;
	int n, nread, nwritten;
	int alpine_fd, retval = -1;

	/* Is it one of ours? */
	alpine_fd = fd_map_lookup_native_fd(native_fd);
	if (alpine_fd == -1) {
		return (native_recvmsg(native_fd, msg, flags));
	}

	bzero(&recvmsg_req, sizeof(recvmsg_req));
	recvmsg_req.req_type = REQ_RECVMSG;
	recvmsg_req.fd = alpine_fd;
	recvmsg_req.namelen = (msg->msg_name!=NULL)?msg->msg_namelen:0;
	recvmsg_req.controllen = (msg->msg_control!=NULL)?msg->msg_controllen:0;
	recvmsg_req.buflen = 0;
	for (n = 0; n < msg->msg_iovlen; ++n) {
		if ((msg->msg_iov[n].iov_base != NULL) &&
		    (msg->msg_iov[n].iov_len > 0)) {
			recvmsg_req.buflen += msg->msg_iov[n].iov_len;
		}
	}
	/* 
	 * XXX: The msg_flags is filled in by the kernel. When sending
	 * data to the kernel we use the 'flags' argument to this function.
	 */
	//recvmsg_req.flags = linux_to_bsd_msg_flags(flags);
	recvmsg_req.flags = flags;

	/* Write out the request */
	nwritten = write_n_bytes(connect_fd, 
				 (char *)&recvmsg_req, sizeof(recvmsg_req));
	if (nwritten != sizeof(recvmsg_req)) {
		CLEANUP_LIB_AND_RETURN();
	}

	/* Read the response */
	nread = read_n_bytes(connect_fd,
			     (char *)&recvmsg_resp, sizeof(recvmsg_resp));
	if (nread != sizeof(recvmsg_resp)) {
		CLEANUP_LIB_AND_RETURN();
	}
	assert(recvmsg_resp.resp_type == RESP_RECVMSG);

	/* And the msg_name if appropriate */
	if (recvmsg_resp.namelen > 0) {
		assert(msg->msg_name != NULL);
		nread = read_n_bytes(connect_fd, 
				     msg->msg_name, recvmsg_resp.namelen);
		if (nread != recvmsg_resp.namelen) {
			CLEANUP_LIB_AND_RETURN();
		}
	}
	msg->msg_namelen = recvmsg_resp.namelen;

	/* And the msg_controllen if appropriate */
	if (recvmsg_resp.controllen > 0) {
		assert(msg->msg_control != NULL);
		nread = read_n_bytes(connect_fd, 
				     msg->msg_control, recvmsg_resp.controllen);
		if (nread != recvmsg_resp.controllen) {
			CLEANUP_LIB_AND_RETURN();
		}
	}
	msg->msg_controllen = recvmsg_resp.controllen;

	/* And the data if appropriate */
	if (recvmsg_resp.retval > 0) {
		int total_read = 0;
		int can_read;
		for (n = 0; n < msg->msg_iovlen; ++n) {
			if ((msg->msg_iov[n].iov_base == NULL) ||
			    (msg->msg_iov[n].iov_len <= 0)) {
				continue;
			}
			can_read = MIN((recvmsg_resp.retval - total_read),
					msg->msg_iov[n].iov_len);
			nread = read_n_bytes(connect_fd,
					     msg->msg_iov[n].iov_base,can_read);
					     
			if (nread != can_read) {
				CLEANUP_LIB_AND_RETURN();
			}
			total_read += nread;
			if (total_read == recvmsg_resp.retval) {
				break;
			}
		}
		assert(total_read == recvmsg_resp.retval);
	}

	/* Set the flags */
	//msg->msg_flags = bsd_to_linux_msg_flags(recvmsg_resp.flags);
	msg->msg_flags = recvmsg_resp.flags;

	/* Set errno */
	if (recvmsg_resp.retval == -1) {
		errno = map_errno(recvmsg_resp.alpine_errno);
	}

	return (recvmsg_resp.retval);
}

int
sendmsg(int native_fd, const struct msghdr *msg, int flags)
{
	struct sendmsg_req sendmsg_req;
	struct sendmsg_resp sendmsg_resp;
	int n, nread, nwritten;
	int alpine_fd, retval = -1;

	/* Is it one of ours? */
	alpine_fd = fd_map_lookup_native_fd(native_fd);
	if (alpine_fd == -1) {
		return (native_sendmsg(native_fd, msg, flags));
	}

	bzero(&sendmsg_req, sizeof(sendmsg_req));
	sendmsg_req.req_type = REQ_SENDMSG;
	sendmsg_req.fd = alpine_fd;
	sendmsg_req.namelen = (msg->msg_name!=NULL)?msg->msg_namelen:0;
	sendmsg_req.controllen = (msg->msg_control!=NULL)?msg->msg_controllen:0;
	sendmsg_req.buflen = 0;
	for (n = 0; n < msg->msg_iovlen; ++n) {
		if ((msg->msg_iov[n].iov_base != NULL) &&
		    (msg->msg_iov[n].iov_len > 0)) {
			sendmsg_req.buflen += msg->msg_iov[n].iov_len;
		}
	}
	//sendmsg_req.flags = linux_to_bsd_msg_flags(flags);
	sendmsg_req.flags = flags;
	/* 
	 * XXX: According to the man page, the msg->msg_flags member is
	 * only set on a received packet.
	 */

	/* Write out the request */
	nwritten = write_n_bytes(connect_fd, 
				 (char *)&sendmsg_req, sizeof(sendmsg_req));
	if (nwritten != sizeof(sendmsg_req)) {
		CLEANUP_LIB_AND_RETURN();
	}
	/* And the msg->msg_name if appropriate */
	if (sendmsg_req.namelen > 0) {
		assert(msg->msg_name != NULL);
		nwritten = write_n_bytes(connect_fd, 
				         msg->msg_name, sendmsg_req.namelen);
		if (nwritten != sendmsg_req.namelen) {
			CLEANUP_LIB_AND_RETURN();
		}
	}
	/* And the msg_controllen if appropriate */
	if (sendmsg_req.controllen > 0) {
		assert(msg->msg_control != NULL);
		nwritten = read_n_bytes(connect_fd, 
				        msg->msg_control, 
					sendmsg_req.controllen);
		if (nwritten != sendmsg_req.controllen) {
			CLEANUP_LIB_AND_RETURN();
		}
	}
	/* And the data if appropriate */
	if (sendmsg_req.buflen > 0) {
		for (n = 0; n < msg->msg_iovlen; ++n) {
			if ((msg->msg_iov[n].iov_base == NULL) ||
			    (msg->msg_iov[n].iov_len <= 0)) {
				continue;
			}
			nwritten = write_n_bytes(connect_fd,
					         msg->msg_iov[n].iov_base,
						 msg->msg_iov[n].iov_len);
					     
			if (nwritten != msg->msg_iov[n].iov_len) {
				CLEANUP_LIB_AND_RETURN();
			}
		}
	}

	/* Read the response */
	nread = read_n_bytes(connect_fd,
			     (char *)&sendmsg_resp, sizeof(sendmsg_resp));
	if (nread != sizeof(sendmsg_resp)) {
		CLEANUP_LIB_AND_RETURN();
	}
	assert(sendmsg_resp.resp_type == RESP_SENDMSG);

	/* Set errno */
	if (sendmsg_resp.retval == -1) {
		errno = map_errno(sendmsg_resp.alpine_errno);
	}

	return (sendmsg_resp.retval);
}

int
recvfrom_internal(int native_fd, void *buf, size_t len, int flags,
		  struct sockaddr *from, socklen_t *fromlen,
		  enum request_type req, enum response_type resp)
{
	struct recvfrom_req recvfrom_req;
	struct recvfrom_resp recvfrom_resp;
	int nread, nwritten;
	int alpine_fd, retval = -1;

	assert((req == REQ_RECV) || (req == REQ_RECVFROM));
	assert((resp == RESP_RECV) || (resp == RESP_RECVFROM));

	/* Is it one of ours? */
	alpine_fd = fd_map_lookup_native_fd(native_fd);
	if (alpine_fd == -1) {
		if (req == REQ_RECV) {
			return (native_recv(native_fd, buf, len, flags));
		} else {
			return (native_recvfrom(native_fd, buf, len, flags,
						from, fromlen));
		}
	}

	recvfrom_req.req_type = req;
	recvfrom_req.fd = alpine_fd;
	recvfrom_req.len = len;
	//recvfrom_req.flags = linux_to_bsd_msg_flags(flags);
	recvfrom_req.flags = flags;
	recvfrom_req.fromlen = ((from != NULL) && (fromlen != NULL)) ?
		 		((*fromlen < 0) ? 0 : *fromlen) : 0;

	/* Write out the request */
	nwritten = write_n_bytes(connect_fd, 
				 (char *)&recvfrom_req, sizeof(recvfrom_req));
	if (nwritten != sizeof(recvfrom_req)) {
		CLEANUP_LIB_AND_RETURN();
	}

	/* Read the response */
	nread = read_n_bytes(connect_fd,
			     (char *)&recvfrom_resp, sizeof(recvfrom_resp));
	if (nread != sizeof(recvfrom_resp)) {
		CLEANUP_LIB_AND_RETURN();
	}
	assert(recvfrom_resp.resp_type == resp);
	assert(recvfrom_resp.fromlen >= 0);

	/* And the data if appropriate */
	if (recvfrom_resp.retval > 0) {
		assert(buf != NULL);
		nread = read_n_bytes(connect_fd, buf, recvfrom_resp.retval);
		if (nread != recvfrom_resp.retval) {
			CLEANUP_LIB_AND_RETURN();
		}
	}

	/* And 'from' if appropriate */
	if (recvfrom_resp.fromlen) {
		assert(from != NULL);
		nread = read_n_bytes(connect_fd, 
				     (char *)from, recvfrom_resp.fromlen);
		if (nread != recvfrom_resp.fromlen) {
			CLEANUP_LIB_AND_RETURN();
		}
	}
	if (fromlen) {
		*fromlen = recvfrom_resp.fromlen;
	}

	/* Set errno */
	if (recvfrom_resp.retval == -1) {
		errno = map_errno(recvfrom_resp.alpine_errno);
	}

#if 0
	{ int i;
		printf("recv_int\n");
		printf("len=%d\n",recvfrom_resp.retval);
		for (i=0;i<recvfrom_resp.retval;i++)
			printf("%x:",*(((char *)buf)+i));
		printf("\n");
		if(from != NULL) {
			printf("addrlen=%d\n",*fromlen);
			for (i=0;i<*fromlen;i++)
				printf("%x:",*(((char *)from)+i));
			printf("\n");
		}
	}
#endif


	return (recvfrom_resp.retval);
}

int
send(int s, const void *buf, size_t len, int flags)
{
	if ((buf == NULL) || (len < 0)) {
		errno = EINVAL;
		return (-1);
	}

	return (sendto_internal(s, buf, len, flags, NULL, 0,
				REQ_SEND, RESP_SEND));
}

int
sendto(int s, const void *buf, size_t len, int flags,
       const struct sockaddr *to, socklen_t tolen)
{
	if ((buf == NULL) || (len < 0)) {
		errno = EINVAL;
		return (-1);
	}

	return (sendto_internal(s, buf, len, flags, to, tolen,
				REQ_SENDTO, RESP_SENDTO));
}

int
sendto_internal(int native_fd, const void *buf, size_t len, int flags,
		const struct sockaddr *to, int tolen,
		enum request_type req, enum response_type resp)
{
	struct sendto_req sendto_req;
	struct sendto_resp sendto_resp;
	int nread, nwritten;
	int alpine_fd, retval = -1;

	assert((req == REQ_SEND) || (req == REQ_SENDTO));
	assert((resp == RESP_SEND) || (resp == RESP_SENDTO));

	/* Is it one of ours? */
	alpine_fd = fd_map_lookup_native_fd(native_fd);
	if (alpine_fd == -1) {
		if (req == REQ_SEND) {
			return (native_send(native_fd, buf, len, flags));
		} else {
			return (native_sendto(native_fd, buf, len, flags,
					      to, tolen));
		}
	}

	sendto_req.req_type = req;
	sendto_req.fd = alpine_fd;
	sendto_req.len = len;
	//sendto_req.flags = linux_to_bsd_msg_flags(flags);
	sendto_req.flags = flags;
	sendto_req.tolen = ((to != NULL) && (tolen > 0)) ?
		 	     tolen : 0;

	/* Write out the request */
	nwritten = write_n_bytes(connect_fd, 
				 (char *)&sendto_req, sizeof(sendto_req));
	if (nwritten != sizeof(sendto_req)) {
		CLEANUP_LIB_AND_RETURN();
	}
	/* And the data if appropriate */
	if (sendto_req.len > 0) {
		assert(buf != NULL);	/* XXX: Return EINVAL instead */
		nwritten = write_n_bytes(connect_fd, 
					 buf, sendto_req.len);
		if (nwritten != sendto_req.len) {
			CLEANUP_LIB_AND_RETURN();
		}
	}
	/* And 'to' if appropriate */
	if (sendto_req.tolen) {
		assert(to != NULL);
		nwritten = write_n_bytes(connect_fd, 
				         (char *)to, sendto_req.tolen);
		if (nwritten != sendto_req.tolen) {
			CLEANUP_LIB_AND_RETURN();
		}
	}

	/* Read the response */
	nread = read_n_bytes(connect_fd,
			     (char *)&sendto_resp, sizeof(sendto_resp));
	if (nread != sizeof(sendto_resp)) {
		CLEANUP_LIB_AND_RETURN();
	}
	assert(sendto_resp.resp_type == resp);

	/* Set errno */
	if (sendto_resp.retval == -1) {
		errno = map_errno(sendto_resp.alpine_errno);
	}

	return (sendto_resp.retval);
}

int
write(int native_fd, const void *data, size_t count)
{
	struct write_req write_req;
	struct write_resp write_resp;
	int nread, nwritten;
	int alpine_fd, retval = -1;

	/* Is it one of ours? */
	alpine_fd = fd_map_lookup_native_fd(native_fd);
	if (alpine_fd == -1) {
		return (native_write(native_fd, data, count));
	}

	write_req.req_type = REQ_WRITE;
	write_req.fd = alpine_fd;
	write_req.count = count;

	/* Write out the request */
	nwritten = write_n_bytes(connect_fd, 
				 (char *)&write_req, sizeof(write_req));
	if (nwritten != sizeof(write_req)) {
		CLEANUP_LIB_AND_RETURN();
	}
	/* And write out the data if appropriate */
	if (write_req.count > 0) {
		nwritten = write_n_bytes(connect_fd, data, write_req.count);
		if (nwritten != write_req.count) {
			CLEANUP_LIB_AND_RETURN();
		}
	}

	/* Read the response */
	nread = read_n_bytes(connect_fd,
			     (char *)&write_resp, sizeof(write_resp));
	if (nread != sizeof(write_resp)) {
		CLEANUP_LIB_AND_RETURN();
	}
	assert(write_resp.resp_type == RESP_WRITE);

	/* Set errno */
	if (write_resp.retval == -1) {
		errno = map_errno(write_resp.alpine_errno);
	}

	return (write_resp.retval);
}

int
setsockopt(int native_fd, int level, int optname, 
	   const void *optval, socklen_t optlen)
{
	struct setsockopt_req setsockopt_req;
	struct setsockopt_resp setsockopt_resp;
	int nread, nwritten;
	int alpine_fd, retval = -1;

	/* Is it one of ours? */
	alpine_fd = fd_map_lookup_native_fd(native_fd);
	if (alpine_fd == -1) {
		return (native_setsockopt(native_fd, level, optname,
					  optval, optlen));
	}

	setsockopt_req.req_type = REQ_SETSOCKOPT;
	setsockopt_req.fd = alpine_fd;
	setsockopt_req.level = level;

	switch (level) {
		case SOL_SOCKET:
		case IPPROTO_IP:
		case IPPROTO_TCP:
		case IPPROTO_IPV6:
		case IPPROTO_ICMPV6:
		case IPPROTO_RAW:
			break;
		default:
			optname = -1;
			break;
	}
	if (optname == -1) {
		errno = EINVAL;
		return (-1);
	}

	setsockopt_req.optname = optname;
	setsockopt_req.optlen = ((optval != NULL) && (optlen > 0)) ?
				 optlen : 0;

	/* write out the request */
	nwritten = write_n_bytes(connect_fd, 
			(char *)&setsockopt_req, sizeof(setsockopt_req));
	if (nwritten != sizeof(setsockopt_req)) {
		CLEANUP_LIB_AND_RETURN();
	}
	/* And write out the optval if appropriate */
	if (setsockopt_req.optlen > 0) {
		nwritten = write_n_bytes(connect_fd, 
					 optval, setsockopt_req.optlen);
		if (nwritten != setsockopt_req.optlen) {
			CLEANUP_LIB_AND_RETURN();
		}
	}

	/* Read the response */
	nread = read_n_bytes(connect_fd,
			     (char *)&setsockopt_resp, sizeof(setsockopt_resp));
	if (nread != sizeof(setsockopt_resp)) {
		CLEANUP_LIB_AND_RETURN();
	}
	assert(setsockopt_resp.resp_type == RESP_SETSOCKOPT);

	/* Set errno */
	if (setsockopt_resp.retval == -1) {
		errno = map_errno(setsockopt_resp.alpine_errno);
	}

	return (setsockopt_resp.retval);
}

int
getsockopt(int native_fd, int level, int optname, 
	   void *optval, socklen_t *optlen)
{
	struct getsockopt_req getsockopt_req;
	struct getsockopt_resp getsockopt_resp;
	int nread, nwritten;
	int alpine_fd, retval = -1;
	int saved_optname = optname;

	/* Is it one of ours? */
	alpine_fd = fd_map_lookup_native_fd(native_fd);
	if (alpine_fd == -1) {
		return (native_getsockopt(native_fd, level, optname,
					  optval, optlen));
	}

	getsockopt_req.req_type = REQ_GETSOCKOPT;
	getsockopt_req.fd = alpine_fd;
	getsockopt_req.level = level;

	switch (level) {
		case SOL_SOCKET:
		case IPPROTO_IP:
		case IPPROTO_TCP:
		case IPPROTO_IPV6:
		case IPPROTO_ICMPV6:
		case IPPROTO_RAW:
			break;
		default:
			optname = -1;
			break;
	}
	if (optname == -1) {
		errno = EINVAL;
		return (-1);
	}

	getsockopt_req.optname = optname;
	getsockopt_req.optlen = ((optval != NULL) && (optlen != NULL)) ?
				 (*optlen > 0 ? *optlen : 0) : 0;

	/* write out the request */
	nwritten = write_n_bytes(connect_fd, 
			(char *)&getsockopt_req, sizeof(getsockopt_req));
	if (nwritten != sizeof(getsockopt_req)) {
		CLEANUP_LIB_AND_RETURN();
	}
	/* And write out the optval if appropriate */
	if (getsockopt_req.optlen > 0) {
		nwritten = write_n_bytes(connect_fd, 
					 optval, getsockopt_req.optlen);
		if (nwritten != getsockopt_req.optlen) {
			CLEANUP_LIB_AND_RETURN();
		}
	}

	/* Read the response */
	nread = read_n_bytes(connect_fd,
			     (char *)&getsockopt_resp, sizeof(getsockopt_resp));
	if (nread != sizeof(getsockopt_resp)) {
		CLEANUP_LIB_AND_RETURN();
	}
	assert(getsockopt_resp.resp_type == RESP_GETSOCKOPT);
	assert(getsockopt_resp.optlen >= 0);

	/* And the optval if appropriate */
	if (getsockopt_resp.optlen > 0) {
		nread = read_n_bytes(connect_fd, 
				     optval, getsockopt_resp.optlen);
		if (nread != getsockopt_resp.optlen) {
			CLEANUP_LIB_AND_RETURN();
		}
	}
	if (optlen != NULL) {
		*optlen = getsockopt_resp.optlen;
	}

	/* XXX */
	if ((level == SOL_SOCKET) && 
	    (saved_optname == SO_ERROR) &&
	    (optval != NULL)) {
		int so_error = *(int *)optval;
		if (so_error) {
			so_error = map_errno(so_error);
		}
		*(int *)optval = so_error;
	}

	/* get errno */
	if (getsockopt_resp.retval == -1) {
		errno = map_errno(getsockopt_resp.alpine_errno);
	}

	return (getsockopt_resp.retval);
}

int
accept(int native_fd, struct sockaddr *addr, socklen_t *ret_addrlen)
{
	struct accept_req accept_req;
	struct accept_resp accept_resp;
	int nread, nwritten;
	int alpine_fd, retval = -1;

#if 0
	if ((addr == NULL) || (ret_addrlen == NULL)) {
		errno = EINVAL;
		return (-1);
	}
#endif

	/* Is it one of ours? */
	alpine_fd = fd_map_lookup_native_fd(native_fd);
	if (alpine_fd == -1) {
		return (native_accept(native_fd, addr, ret_addrlen));
	}

	accept_req.req_type = REQ_ACCEPT;
	accept_req.fd = alpine_fd;
	accept_req.addrlen = ((addr != NULL) && (ret_addrlen != NULL)) ? 
		              ((*ret_addrlen < 0) ? 0 : *ret_addrlen) : 0;

	/* Write out the request */
	nwritten = write_n_bytes(connect_fd, 
				 (char *)&accept_req, sizeof(accept_req));
	if (nwritten != sizeof(accept_req)) {
		CLEANUP_LIB_AND_RETURN();
	}

	/* And the addr if appropriate */
	if (accept_req.addrlen) {
		nwritten = write_n_bytes(connect_fd, 
					 (char *)addr, accept_req.addrlen);
		if (nwritten != accept_req.addrlen) {
			CLEANUP_LIB_AND_RETURN();
		}
	}

	/* Read the response */
	nread = read_n_bytes(connect_fd,
			     (char *)&accept_resp, sizeof(accept_resp));
	if (nread != sizeof(accept_resp)) {
		CLEANUP_LIB_AND_RETURN();
	}
	assert(accept_resp.resp_type == RESP_ACCEPT);
	assert(accept_resp.addrlen >= 0);

	/* And the addr if appropriate */
	if (accept_resp.addrlen) {
		nread = read_n_bytes(connect_fd,
				     (char *)addr, accept_resp.addrlen);
		if (nread != accept_resp.addrlen) {
			CLEANUP_LIB_AND_RETURN();
		}
	}
	if (ret_addrlen) {
		*ret_addrlen = accept_resp.addrlen;
	}

	/* Set errno */
	if (accept_resp.retval == -1) {
		errno = map_errno(accept_resp.alpine_errno);
		return (-1);
	} else {
		int new_fd;
		new_fd = get_native_fd();
		if (new_fd == -1) {
			close(accept_resp.retval); /* Close the stack fd */
			return (-1);
		}
		fd_map_create(new_fd, accept_resp.retval);
		return (new_fd);
	}

	assert(0);
}

int 
getsockname(int native_fd, struct sockaddr *name, socklen_t *ret_namelen)
{
	return (getname(native_fd, name, ret_namelen, 
			REQ_GETSOCKNAME, RESP_GETSOCKNAME));
}

int 
getpeername(int native_fd, struct sockaddr *name, socklen_t *ret_namelen)
{
	return (getname(native_fd, name, ret_namelen, 
			REQ_GETPEERNAME, RESP_GETPEERNAME));
}

int
getname(int native_fd, struct sockaddr *name, socklen_t *ret_namelen,
	enum request_type req, enum response_type resp)
{
	struct getname_req getname_req;
	struct getname_resp getname_resp;
	int nread, nwritten;
	int alpine_fd, retval = -1;

	/* XXX */
#if 0	
	if ((name == NULL) || (ret_namelen == NULL)) {
		errno = EINVAL;
		return (-1);
	}
#endif

	/* Is it one of ours? */
	alpine_fd = fd_map_lookup_native_fd(native_fd);
	if (alpine_fd == -1) {
		if (req == REQ_GETSOCKNAME) {
			return (native_getsockname(native_fd,name,ret_namelen));
		} else {
			return (native_getpeername(native_fd,name,ret_namelen));
		}
	}

	getname_req.req_type = req;
	getname_req.fd = alpine_fd;
	getname_req.namelen = ((name != NULL) && (ret_namelen != NULL)) ? 
		              ((*ret_namelen < 0) ? 0 : *ret_namelen) : 0;

	/* Write out the request */
	nwritten = write_n_bytes(connect_fd, 
				 (char *)&getname_req, sizeof(getname_req));
	if (nwritten != sizeof(getname_req)) {
		CLEANUP_LIB_AND_RETURN();
	}

	/* Read the response */
	nread = read_n_bytes(connect_fd,
			     (char *)&getname_resp, sizeof(getname_resp));
	if (nread != sizeof(getname_resp)) {
		CLEANUP_LIB_AND_RETURN();
	}
	assert(getname_resp.resp_type == resp);
	assert(getname_resp.namelen >= 0);

	/* And the name if appropriate */
	if (getname_resp.namelen) {
		assert(name != NULL);
		nread = read_n_bytes(connect_fd,
				     (char *)name, getname_resp.namelen);
		if (nread != getname_resp.namelen) {
			CLEANUP_LIB_AND_RETURN();
		}
	}
	if (ret_namelen) {
		*ret_namelen = getname_resp.namelen;
	}

	/* Set errno */
	if (getname_resp.retval == -1) {
		errno = map_errno(getname_resp.alpine_errno);
	}

	return (getname_resp.retval);
}

int
ioctl(int native_fd, unsigned long int command, ...)
{
	struct ioctl_req ioctl_req;
	struct ioctl_resp ioctl_resp;
	int nread, nwritten;
	int data_length;
	char *data = NULL;
	int alpine_fd, retval = -1;
	va_list ap;
	struct ifconf *ifc, saved_ifconf;

	va_start(ap, command);
	data = va_arg(ap, char *);
	va_end(ap);

	/* Is it one of ours? */
	alpine_fd = fd_map_lookup_native_fd(native_fd);
	if (alpine_fd == -1) {
		return (native_ioctl(native_fd, command, data));
	}

	switch (command) {
		/* Routing table calls */
		case SIOCADDRT:
		case SIOCDELRT:
		case SIOCRTMSG:
			data_length = -1;
			break;
		
		/* Socket configuration controls */
		case SIOCGIFNAME:
		case SIOCSIFLINK:
			data_length = -1;
			break;

		case SIOCGIFCONF:
			bzero(&saved_ifconf, sizeof(saved_ifconf));
			data_length = sizeof(struct ifconf);
			if (data != NULL) {
				saved_ifconf = *(struct ifconf *)data;
			} 
			break;

		case SIOCGIFFLAGS:
		case SIOCSIFFLAGS:
		case SIOCGIFHWADDR:
		case SIOCGIFMETRIC:
		case SIOCSIFMETRIC:
		case SIOCGIFMTU:
		case SIOCSIFMTU:
		case SIOCSIFNETMASK:
		case SIOCGIFNETMASK:

		case SIOCGIFADDR:
		case SIOCGIFDSTADDR:
		case SIOCGIFBRDADDR:

		case SIOCSIFADDR:
		case SIOCSIFDSTADDR:
		case SIOCSIFBRDADDR:

		case SIOCDIFADDR:
			data_length = sizeof(struct ifreq);
			break;

		case SIOCSIFHWADDR:
			data_length = sizeof(struct ifreq);
			break;

		case SIOCSIFTXQLEN:
		case SIOCSIFMAP:
			/* XXX: Does FreeBSD support this? */
			data_length = -1;
			break;

		case FIONREAD:
		case FIONBIO:
			data_length = sizeof(int);
			break;

		default:
			data_length = -1;
			break;
	}

	/* Validate the command and data */
	if (data_length == -1) {
		errno = EINVAL;	/* XXX */
		return (-1);
	} else if (data_length > 0) {
		if (data == NULL) {
			errno = EINVAL;
			return (-1);
		}
	} else {
		assert(data_length == 0);
	}

	ioctl_req.req_type = REQ_IOCTL;
	ioctl_req.fd = alpine_fd;
	ioctl_req.command = command;
	ioctl_req.data_length = data_length;

	/* Write out the request */
	nwritten = write_n_bytes(connect_fd, 
				 (char *)&ioctl_req, sizeof(ioctl_req));
	if (nwritten != sizeof(ioctl_req)) {
		CLEANUP_LIB_AND_RETURN();
	}

	/* And the additional data if necessary */
	if (data_length) {
		nwritten = write_n_bytes(connect_fd, (char *)data, data_length);
		if (nwritten != data_length) {
			CLEANUP_LIB_AND_RETURN();
		}
	}

	/* Read the response */
	nread = read_n_bytes(connect_fd,
			     (char *)&ioctl_resp, sizeof(ioctl_resp));
	if (nread != sizeof(ioctl_resp)) {
		CLEANUP_LIB_AND_RETURN();
	}
	assert(ioctl_resp.resp_type == RESP_IOCTL);

	/* And the data if necessary */
	if (data_length) {
		nread = read_n_bytes(connect_fd, data, data_length);
		if (nread != data_length) {
			CLEANUP_LIB_AND_RETURN();
		}
	}

	switch (command) {
		case SIOCGIFCONF:
			ifc = (struct ifconf *)data;
			ifc->ifc_buf = saved_ifconf.ifc_buf;
			if (ifc->ifc_len > 0) {
				nread = read_n_bytes(connect_fd, ifc->ifc_buf,
						     ifc->ifc_len);
				if (nread != ifc->ifc_len) {
					CLEANUP_LIB_AND_RETURN();
				}
			}
			break;
		
		default:
			break;
	}

	/* Set errno */
	if (ioctl_resp.retval == -1) {
		errno = map_errno(ioctl_resp.alpine_errno);
	}

	return (ioctl_resp.retval);
}

int
fcntl(int native_fd, int cmd, ...)
{
	struct fcntl_req fcntl_req;
	struct fcntl_resp fcntl_resp;
	int nread, nwritten;
	int alpine_fd;
	va_list ap;
	long arg;

	va_start(ap, cmd);
	arg = va_arg(ap, long);
	va_end(ap);

	alpine_fd = fd_map_lookup_native_fd(native_fd);
	if (alpine_fd == -1) {
		return (native_fcntl(native_fd, cmd, arg));
	}

	/* Send the request */
	fcntl_req.req_type = REQ_FCNTL;
	fcntl_req.fd = alpine_fd;

	switch (cmd) {
	case F_GETFD:
		fcntl_req.cmd = F_GETFD;
		fcntl_req.arg = 0; /* unused */
		break;
	
	case F_SETFD:
		fcntl_req.cmd = F_SETFD;
		fcntl_req.arg = arg;
		break;

	case F_GETFL:
		fcntl_req.cmd = F_GETFL;
		fcntl_req.arg = 0; /* unused */
		break;
	
	case F_SETFL:
		fcntl_req.cmd = F_SETFL;
		fcntl_req.arg = arg;
		break;
	
	default:
		errno = EINVAL;
		return (-1);
	}

	nwritten = write_n_bytes(connect_fd,
				 (char *)&fcntl_req, 
				 sizeof(fcntl_req));
	if (nwritten != sizeof(fcntl_req)) {
		CLEANUP_LIB_AND_RETURN();
	}

	/* Read the response */
	nread = read_n_bytes(connect_fd,
			     (char *)&fcntl_resp,
			     sizeof(fcntl_resp));
	if (nread != sizeof(fcntl_resp)) {
		CLEANUP_LIB_AND_RETURN();
	}
	assert(fcntl_resp.resp_type == RESP_FCNTL);
	if (fcntl_resp.retval == -1) {
		errno = map_errno(fcntl_resp.alpine_errno);
	} else {
		switch (cmd) {
		case F_GETFL:
			arg = fcntl_resp.retval;
			/*if (fcntl_resp.retval & FREEBSD_O_RDONLY)
				arg |= O_RDONLY;
			if (fcntl_resp.retval & FREEBSD_O_WRONLY)
				arg |= O_WRONLY;
			if (fcntl_resp.retval & FREEBSD_O_RDWR)
				arg |= O_RDWR;
			if (fcntl_resp.retval & FREEBSD_O_NONBLOCK)
				arg |= O_NONBLOCK;
			if (fcntl_resp.retval & FREEBSD_O_APPEND)
				arg |= O_APPEND;
			if (fcntl_resp.retval & FREEBSD_O_FSYNC)
				arg |= O_SYNC;
			if (fcntl_resp.retval & FREEBSD_O_ASYNC)
				arg |= FASYNC;
			fcntl_resp.retval = arg;*/
			break;

		case F_SETFL:
			/* On Linux fcntl(F_SETFL) return 0 on success */
			if (fcntl_resp.retval != -1) {
				fcntl_resp.retval = 0;
			}
			break;

		default:
			break;
		}
	}

	return (fcntl_resp.retval);
}

int
select(int nfd, fd_set *native_rdset, fd_set *native_wrset,
       fd_set *native_exset, struct timeval *tv)
{
	int nread, nwritten, saved_errno;
	int n, retval, nset, alpine_nfd;
	struct select_req select_req;
	struct select_stop_req select_stop_req;
	struct select_resp select_resp;
	fd_set alpine_rdset;
	fd_set alpine_wrset;
	fd_set alpine_exset;
	fd_set our_native_rdset;
	fd_set *alpine_sets[3];

	alpine_sets[0] = &alpine_rdset;
	alpine_sets[1] = &alpine_wrset;
	alpine_sets[2] = &alpine_exset;

	INIT_LIB();

	/*
	 * The caller might not specify a native_rdset. But *we* may
	 * require it if we need to select() on 'connect_fd'. So we
	 * provide one if the caller has not supplied one already.
	 */
	FD_ZERO(&our_native_rdset);
	if (native_rdset == NULL) {
		native_rdset = &our_native_rdset;
	}

	/* 
	 * No reason why the caller should have 'connect_fd' in her
	 * native_rdset. We own 'connect_fd'.
	 */
	if ((native_rdset && FD_ISSET(connect_fd, native_rdset)) ||
	    (native_wrset && FD_ISSET(connect_fd, native_wrset)) ||
	    (native_exset && FD_ISSET(connect_fd, native_exset))) {
		errno = EBADF;
		return (-1);
	}

	alpine_nfd = populate_alpine_fdset(nfd, 
				native_rdset, native_wrset, native_exset,
				&alpine_rdset, &alpine_wrset, &alpine_exset);
	alpine_nfd = alpine_nfd + 1;	/* 1 greater than the max_alpine_fd */
	assert((alpine_nfd >= 0) && (alpine_nfd <= FD_SETSIZE));

	if (alpine_nfd == 0) {
		/* Just do a native_select() */	
		return (native_select(nfd, native_rdset, native_wrset,
				      native_exset, tv));
	}

	/* 
	 * We found at least one alpine_fd we want to select() on,
	 * then put the 'connect_fd' in the native_rdset, so that we
	 * wakeup when the alpine_server sends us the select_resp.
	 */
	assert(connect_fd >= 0);
	FD_SET(connect_fd, native_rdset);
	/* Adjust 'nfd' for the 'connect_fd' we just inserted in the rdset */
	if (nfd <= connect_fd) {
		nfd = connect_fd + 1;
	}

	select_req.req_type = REQ_SELECT;
	select_req.nfd = alpine_nfd;

	/* Write the select_req */
	nwritten = write_n_bytes(connect_fd,
				 (const char *)&select_req, 
				 sizeof(select_req));
	if (nwritten != sizeof(select_req)) {
		CLEANUP_LIB_AND_RETURN();
	}
	/* Now write out the fd_mask arrays */
	for (nset = 0; nset < 3; ++nset) {
		int nbytes = howmany(alpine_nfd, NFDBITS)*sizeof(fd_mask);
		nwritten = write_n_bytes(connect_fd,
			(const char *)__FDS_BITS(alpine_sets[nset]),
			nbytes);
		if (nwritten != nbytes) {
			CLEANUP_LIB_AND_RETURN();
		}
	}

	/* Now do the native_select() */
	n = native_select(nfd, native_rdset, native_wrset, native_exset, tv);
	saved_errno = errno;

	/*
	 * 1. The alpine_server sends us a select_resp
	 * 2. Error (asynchronous like EINTR or synchronous like EBADF)
	 * 3. Timeout
	 * 4. One of the native_fds is activated
	 */

	if ((n == -1) || 				/* Case 2 */
	    (n == 0) ||					/* Case 3 */
	    (!FD_ISSET(connect_fd, native_rdset))) {	/* Case 4 */
		select_stop_req.req_type = REQ_SELECT_STOP;
		nwritten = write_n_bytes(connect_fd,
					 (const char *)&select_stop_req,
					 sizeof(select_stop_req));
		if (nwritten != sizeof(select_stop_req)) {
			CLEANUP_LIB_AND_RETURN();
		}
	}

	/* 
	 * Now we wait synchronously for a select_resp from the server.
	 * We will get this response because the alpine_server already
	 * sent it to us (case 1) or we had to trigger it explicitly 
	 * (case 2,3,4).
	 */
	nread = read_n_bytes(connect_fd, 
			     (char *)&select_resp, sizeof(select_resp));
	if (nread != sizeof(select_resp)) {
		CLEANUP_LIB_AND_RETURN();
	}
	assert(select_resp.resp_type == RESP_SELECT);

	/* Now read the fd_mask arrays */
	for (nset = 0; nset < 3; ++nset) {
		int nbytes = howmany(alpine_nfd, NFDBITS)*sizeof(fd_mask);
		nread = read_n_bytes(connect_fd,
				(char *)__FDS_BITS(alpine_sets[nset]),
				nbytes);
		if (nread != nbytes) {
			CLEANUP_LIB_AND_RETURN();
		}
	}

	/* 
	 * The alpine_server sent us a real response (in contrast to
	 * a response forced due to cancelling select()). Reflect the
	 * returned alpine_sets[] to native_sets[].
	 */
	if ((n > 0) && (FD_ISSET(connect_fd, native_rdset))) {
		/* 
		 * Account for the activation of connect_fd, that
		 * the caller knows nothing about.
		 */
		--n;
		FD_CLR(connect_fd, native_rdset);
	}

	if (select_resp.retval == -1) {
		retval = -1;
		errno = map_errno(select_resp.alpine_errno);
	} else if (select_resp.retval > 0) {
		retval = populate_native_fdset(alpine_nfd,
			&alpine_rdset, &alpine_wrset, &alpine_exset,
			native_rdset, native_wrset, native_exset);
		assert(retval > 0);
		retval += n;
	} else {
		/* 
		 * select_resp.retval == 0 can happen when we had
		 * to forcefully cancel the freebsd_select(). In that
		 * case, use the original errno and retval.
		 */
		errno = saved_errno;
		retval = n;
	}

	return (retval);
}

int
populate_alpine_fdset(int nfd, 
	fd_set *native_rdset, fd_set *native_wrset, fd_set *native_exset,
	fd_set *alpine_rdset, fd_set *alpine_wrset, fd_set *alpine_exset)
{
	int nset, alpine_fd, native_fd;
	int max_alpine_fd = -1;
	fd_set *native_sets[3], *alpine_sets[3];

	assert(alpine_rdset && alpine_wrset && alpine_exset);

	native_sets[0] = native_rdset;
	native_sets[1] = native_wrset;
	native_sets[2] = native_exset;

	alpine_sets[0] = alpine_rdset;
	alpine_sets[1] = alpine_wrset;
	alpine_sets[2] = alpine_exset;

	for (nset = 0; nset < 3; ++nset) {
		FD_ZERO(alpine_sets[nset]);
		if (native_sets[nset] == NULL) {
			continue;
		}
		for (native_fd = 0; native_fd < nfd; ++native_fd) {
			if (!FD_ISSET(native_fd, native_sets[nset])) {
				continue;
			}
			alpine_fd = fd_map_lookup_native_fd(native_fd);
			if (alpine_fd == -1) {
				continue;
			}
			/* One of ours */
			assert(alpine_fd < FD_SETSIZE);	/* XXX */
			FD_CLR(native_fd, native_sets[nset]);
			FD_SET(alpine_fd, alpine_sets[nset]);
			if (alpine_fd > max_alpine_fd) {
				max_alpine_fd = alpine_fd;
			}
		}
	}

	return (max_alpine_fd);
}

int
populate_native_fdset(int alpine_nfd,
		const fd_set *alpine_rdset, 
		const fd_set *alpine_wrset, 
		const fd_set *alpine_exset,
		fd_set *native_rdset, 
		fd_set *native_wrset, 
		fd_set *native_exset)
{
	int nset, alpine_fd, native_fd;
	int num_native_fds = 0;
	fd_set *native_sets[3];
	const fd_set *alpine_sets[3];

	assert(alpine_rdset && alpine_wrset && alpine_exset);

	native_sets[0] = native_rdset;
	native_sets[1] = native_wrset;
	native_sets[2] = native_exset;

	alpine_sets[0] = alpine_rdset;
	alpine_sets[1] = alpine_wrset;
	alpine_sets[2] = alpine_exset;

	for (nset = 0; nset < 3; ++nset) {
		if (native_sets[nset] == NULL) {
			continue;
			/*
			 * XXX:
			 * At this point we could assert that
			 * alpine_sets[nset] does not contain any
			 * bit that is set
			 */
		}
		for (alpine_fd = 0; alpine_fd < alpine_nfd; ++alpine_fd) {
			if (!FD_ISSET(alpine_fd, alpine_sets[nset])) {
				continue;
			}
			native_fd = fd_map_lookup_alpine_fd(alpine_fd);
			assert(native_fd != -1);
			FD_SET(native_fd, native_sets[nset]);
			++num_native_fds;
		}
	}

	return (num_native_fds);
}

pid_t
fork(void)
{
	int child_connect_fd;
	int32_t child_connect_pid;
	struct fork_req fork_req;
	struct fork_resp fork_resp;
	int nread, nwritten;
	pid_t native_pid = -1;

	INIT_LIB();

	assert(connect_fd >= 0);
	assert(connect_pid >= 0);
	assert(connect_addr != NULL);
	//assert(connect_port != -1);

	/* 
	 * First establish a new connection to the alpine_server.
	 * This connection will be used by the child after the fork.
	 */
	//child_connect_fd = do_connect(connect_addr, connect_port, 
			              //&child_connect_pid);
	child_connect_fd = do_connect(connect_addr, 
			              &child_connect_pid);
	if (child_connect_fd < 0) {
		errno = EAGAIN;
		return (-1);
	}

	fork_req.req_type = REQ_FORK;
	fork_req.parent_pid = connect_pid;
	fork_req.child_pid = child_connect_pid;

	/* Write out the request */
	nwritten = write_n_bytes(connect_fd, 
				 (char *)&fork_req, sizeof(fork_req));
	if (nwritten != sizeof(fork_req)) {
		native_close(child_connect_fd);
		CLEANUP_LIB_AND_RETURN();
	}

	/* Now wait for the response */
	nread = read_n_bytes(connect_fd, 
			     (char *)&fork_resp, sizeof(fork_resp));
	if (nread != sizeof(fork_resp)) {
		native_close(child_connect_fd);
		CLEANUP_LIB_AND_RETURN();
	}
	assert(fork_resp.resp_type == RESP_FORK);

	/* 
	 * The FreeBSD stack had trouble copying file descriptors from the
	 * parent process to the child process. fork() cannot continue.
	 */
	if (fork_resp.retval != 0) {
		native_close(child_connect_fd);
		errno = map_errno(fork_resp.alpine_errno);
		return (-1);
	}
	
	/* Now do the real fork() */
	native_pid = native_fork();

	/* Linux fork() had an error */
	if (native_pid == -1) {
		int saved_errno = errno;
		native_close(child_connect_fd);
		errno = saved_errno;
	}

	/* Child process */
	else if (native_pid == 0) {
		/* We have no use for the parent's connection to the server */
		native_close(connect_fd);
		connect_fd = child_connect_fd;
		connect_pid = child_connect_pid;
	} 

	/* Parent process */
	else {
		/* We have no use for the child's connection to the server */
		native_close(child_connect_fd);
	}

	return (native_pid);
}

int
dup2(int old_native_fd, int new_native_fd)
{
	int rc, old_alpine_fd, new_alpine_fd;

	/* dup2() of the old_native_fd to itself */
	if (old_native_fd == new_native_fd) {
		return (new_native_fd);
	}

	old_alpine_fd = fd_map_lookup_native_fd(old_native_fd);
	new_alpine_fd = fd_map_lookup_native_fd(new_native_fd);

	/*
	 * The source fd does not belong to Alpine, so we simply call
	 * into the native_dup2(). Of course, we need to be careful to
	 * close the new_native_fd, if it earlier belonged to Alpine.
	 */
	if (old_alpine_fd == -1) {
		if (new_alpine_fd != -1) {
			close(new_native_fd);
		}
		return (native_dup2(old_native_fd, new_native_fd));
	} 
	/*
	 * The old_native_fd belongs to Alpine therefore we must do
	 * the mapping in two places. Once in the Linux kernel and once
	 * in Alpine.
	 */
	else {
		if (new_alpine_fd != -1) {
			close(new_native_fd);
		}

		/*
		 * We just closed new_native_fd, so there is no reason
		 * why the native_dup2() should fail.
		 */
		rc = native_dup2(old_native_fd, new_native_fd);
		assert(rc == new_native_fd);

		/* 
		 * assert() on a successful dup(). This is because there
		 * is no way for us to "go back"; as we might have already
		 * closed new_native_fd.
		 */
		new_alpine_fd = freebsd_dup(old_alpine_fd);
		assert(new_alpine_fd >= 0);

		fd_map_create(new_native_fd, new_alpine_fd);
		return (new_native_fd);
	}

	assert(0);
	return (-1);
}

int
dup(int old_native_fd)
{
	int old_alpine_fd;
	int new_alpine_fd, new_native_fd;

	old_alpine_fd = fd_map_lookup_native_fd(old_native_fd);
	if (old_alpine_fd == -1) {
		return (native_dup(old_native_fd));
	} else {
		new_native_fd = native_dup(old_native_fd);
		if (new_native_fd == -1) {
			return (-1);
		}

		new_alpine_fd = freebsd_dup(old_alpine_fd);
		if (new_alpine_fd < 0) {
			native_close(new_native_fd);
			return (-1);
		}
		
		/* Put them together */
		fd_map_create(new_native_fd, new_alpine_fd);
		return (new_native_fd);
	}

	assert(0);
}

int
freebsd_dup(int old_fd)
{
	int nread, nwritten;
	struct dup_req dup_req;
	struct dup_resp dup_resp;

	INIT_LIB();

	assert(old_fd >= 0);

	dup_req.req_type = REQ_DUP;
	dup_req.old_fd = old_fd;
	dup_req.new_fd = -1;

	/* Write out the request */
	nwritten = write_n_bytes(connect_fd, 
				 (char *)&dup_req, sizeof(dup_req));
	if (nwritten != sizeof(dup_req)) {
		CLEANUP_LIB_AND_RETURN();
	}

	/* Now wait for the response */
	nread = read_n_bytes(connect_fd, 
			     (char *)&dup_resp, sizeof(dup_resp));
	if (nread != sizeof(dup_resp)) {
		CLEANUP_LIB_AND_RETURN();
	}
	assert(dup_resp.resp_type == RESP_DUP);
	
	/* Create a mapping between native_fd and dup_fd */
	if (dup_resp.retval >= 0) {
		return (dup_resp.retval);
	} else {
		errno = map_errno(dup_resp.alpine_errno);
		return (-1);
	}

	assert(0);
	return (-1);
}

ssize_t
sendfile(int native_out_fd, int native_in_fd, off_t *offset, size_t count)
{
	int nread, nwritten, total_written = 0;
	unsigned char buf[8192];

	/* Sockets are not allowed as input fds */
	if (fd_map_lookup_native_fd(native_in_fd) != -1) {
		errno = EINVAL;
		return (-1);
	}

	/* 
	 * If the output fd is not owned by Alpine then let the
	 * Linux kernel handle it.
	 */
	if (fd_map_lookup_native_fd(native_out_fd) == -1) {
		return (native_sendfile(native_out_fd, native_in_fd,
					offset, count));
	}

	/* XXX: Verify the input fd is readable using fstat() */

	if (offset) {
		/* 
		 * The native_in_fd is known to be *not* a socket. Hence we can
		 * safely use native_lseek() and native_read() on it.
		 */
		if (native_lseek(native_in_fd, *offset, SEEK_SET) == 
				 (off_t) -1) {
			return (-1);
		}
	}

	while (count > 0) {
		nread = native_read(native_in_fd,buf,MIN(count, sizeof(buf)));
		if (nread == -1) {			/* Error */
			if (total_written == 0) {
				/* Nothing interesting has happened */
				return (-1);
			} else {
				/* Report back whatever we wrote */
				break;
			}
		} else if (nread == 0) { 		/* EOF */
			break;
		} else {
			nwritten = write(native_out_fd, buf, nread);
			if (nwritten == -1) {
				if (total_written == 0) {
					/* Nothing interesting has happened */
					return (-1);
				} else {
					/* Report back whatever we wrote */
					break;
				}
			}
			total_written += nwritten;

			if (nwritten != nread) {
				break;
			}
		}

		assert(nwritten == nread);
		count -= nwritten;
	}

	if (offset) {
		/* XXX: Reset the file pointer to where it was */
		(void) native_lseek(native_in_fd, *offset, SEEK_SET);
		*offset += total_written;
	}

	return (total_written);
}

int __poll(void *x, int y, int z)
{
	return poll(x,y,z);
}

