/*   This is part of ALE4Net
 *   Application Level Environment for Networking
 *
 *   fd_map: mapping between a4n_server fd and native fd.
 *   
 *   Copyright 2004 Renzo Davoli University of Bologna - Italy
 *   
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License along
 *   with this program; if not, write to the Free Software Foundation, Inc.,
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */   
/*
 * Neelkanth Natu (neelnatu@yahoo.com) is the author of this file.
 * As long as you keep this blurb in the file, feel free to do whatever.
 */ 
#include <sys/queue.h>
#include <assert.h>
#include <unistd.h>

#include "fd_map.h"

#ifndef TAILQ_FIRST
#define	TAILQ_FIRST(head)	((head)->tqh_first)
#endif

#ifndef TAILQ_NEXT
#define	TAILQ_NEXT(elm, field) ((elm)->field.tqe_next)
#endif

#ifndef TAILQ_FOREACH
#define	TAILQ_FOREACH(var, head, field)					\
	for ((var) = TAILQ_FIRST((head));				\
	    (var);							\
	    (var) = TAILQ_NEXT((var), field))
#endif

struct fd_map {
	int native_fd;
	int alpine_fd;
	TAILQ_ENTRY(fd_map) fd_map_link;
};
static TAILQ_HEAD(, fd_map) fd_map_head;

void
fd_map_init(void)
{
	TAILQ_INIT(&fd_map_head);
}

void
fd_map_create (int native_fd, int alpine_fd)
{
	struct fd_map *fd_map;

	assert(native_fd >= 0);
	assert(alpine_fd >= 0);
	assert(fd_map_lookup_native_fd(native_fd) == -1);
	assert(fd_map_lookup_alpine_fd(alpine_fd) == -1);

	fd_map = (struct fd_map *)malloc(sizeof(*fd_map));
	assert(fd_map != NULL);		/* XXX */

	bzero(fd_map, sizeof(*fd_map));
	fd_map->native_fd = native_fd;
	fd_map->alpine_fd = alpine_fd;
	TAILQ_INSERT_HEAD(&fd_map_head, fd_map, fd_map_link);

	return;
}

void
fd_map_destroy(int native_fd)
{
	struct fd_map *fd_map;

	TAILQ_FOREACH(fd_map, &fd_map_head, fd_map_link) {
		if (fd_map->native_fd == native_fd) {
			break;
		}
	}
	assert(fd_map != NULL);

	TAILQ_REMOVE(&fd_map_head, fd_map, fd_map_link);

	bzero(fd_map, sizeof(*fd_map));
	free(fd_map);

	return;
}

int
fd_map_lookup_native_fd(int native_fd)
{
	struct fd_map *fd_map;

	TAILQ_FOREACH(fd_map, &fd_map_head, fd_map_link) {
		if (fd_map->native_fd == native_fd) {
			return (fd_map->alpine_fd);
		}
	}

	return (-1);
}

int
fd_map_lookup_alpine_fd(int alpine_fd)
{
	struct fd_map *fd_map;

	TAILQ_FOREACH(fd_map, &fd_map_head, fd_map_link) {
		if (fd_map->alpine_fd == alpine_fd) {
			return (fd_map->native_fd);
		}
	}

	return (-1);
}
