/*   This is part of ALE4Net
 *   Application Level Environment for Networking
 *   
 *   native_syscall: shortcut to the syscalls (libc routines have been 
 *   overwritten by LD_PRELOAD
 *
 *   Copyright 2004 Renzo Davoli University of Bologna - Italy
 *   
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License along
 *   with this program; if not, write to the Free Software Foundation, Inc.,
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */   
/*
 * Neelkanth Natu (neelnatu@yahoo.com) is the author of this file.
 * As long as you keep this blurb in the file, feel free to do whatever.
 */ 
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/syscall.h>
#include <netinet/in.h>
#include <linux/net.h>	/* for SYS_SOCKET, SYS_BIND etc */

int 
native_socket(int domain, int type, int protocol)
{
	unsigned long args[] = {domain, type, protocol};

	return (syscall(SYS_socketcall, SYS_SOCKET, args));
}

int 
native_bind(int sockfd, struct sockaddr *my_addr, socklen_t addrlen)
{
	unsigned long args[] = {sockfd, (unsigned long)my_addr, addrlen};

	return (syscall(SYS_socketcall, SYS_BIND, args));
}

int 
native_connect(int sockfd, const struct sockaddr *servaddr, socklen_t addrlen)
{
	unsigned long args[] = {sockfd, (unsigned long)servaddr, addrlen};

	return (syscall(SYS_socketcall, SYS_CONNECT, args));
}

int 
native_listen(int s, int backlog)
{
	unsigned long args[] = {s, backlog};

	return (syscall(SYS_socketcall, SYS_LISTEN, args));
}

int 
native_getsockname(int s, struct sockaddr *name, socklen_t *namelen)
{
	unsigned long args[] = {
		s, (unsigned long)name, (unsigned long)namelen
	};

	return (syscall(SYS_socketcall, SYS_GETSOCKNAME, args));
}

int 
native_getpeername(int s, struct sockaddr *name, socklen_t *namelen)
{
	unsigned long args[] = {
		s, (unsigned long)name, (unsigned long)namelen
	};

	return (syscall(SYS_socketcall, SYS_GETPEERNAME, args));
}

int 
native_accept(int s, struct sockaddr *my_addr, socklen_t *addrlen)
{
	unsigned long args[] = {
		s, (unsigned long)my_addr, (unsigned long)addrlen
	};

	return (syscall(SYS_socketcall, SYS_ACCEPT, args));
}

int 
native_setsockopt(int fd, int level, int optname, 
	          void *optval, socklen_t optlen)
{
	unsigned long args[] = {
		fd, level, optname, (unsigned long)optval, optlen
	};

	return (syscall(SYS_socketcall, SYS_SETSOCKOPT, args));
}

int 
native_getsockopt(int fd, int level, int optname, 
	          void *optval, socklen_t *optlen)
{
	unsigned long args[] = {
		fd,level,optname,(unsigned long)optval,(unsigned long)optlen
	};

	return (syscall(SYS_socketcall, SYS_GETSOCKOPT, args));
}

int 
native_select(int nfds, fd_set *rdset, fd_set *wrset, fd_set *xpset,
	      struct timeval *timeout)
{
	return (syscall(SYS__newselect, nfds, rdset, wrset, xpset, timeout));
}

int 
native_write(int fd, void *buf, size_t count)
{
	return (syscall(SYS_write, fd, buf, count));
}

int 
native_read(int fd, void *buf, size_t count)
{
	return (syscall(SYS_read, fd, buf, count));
}

int 
native_close(int fd)
{
	return (syscall(SYS_close, fd));
}

int 
native_open(const char *pathname, int flags)
{
	return (syscall(SYS_open, pathname, flags));
}

int
native_ioctl(int fd, unsigned long int command, char *data)
{
	return (syscall(SYS_ioctl, fd, command, data));
}

int 
native_send(int s, const void *msg, size_t len, int flags)
{
	unsigned long args[] = {
		s, (unsigned long)msg, len, flags
	};

	return (syscall(SYS_socketcall, SYS_SEND, args));
}

int 
native_sendto(int s, const void *msg, size_t len, int flags,
	      struct sockaddr *to, socklen_t tolen)
{
	unsigned long args[] = {
		s, (unsigned long)msg, len, flags, (unsigned long)to, tolen
	};

	return (syscall(SYS_socketcall, SYS_SENDTO, args));
}

int 
native_sendmsg(int s, const struct msghdr *msg, int flags)
{
	unsigned long args[] = {
		s, (unsigned long)msg, flags
	};

	return (syscall(SYS_socketcall, SYS_SENDMSG, args));
}

int 
native_recv(int s, void *buf, size_t len, int flags)
{
	unsigned long args[] = {
		s, (unsigned long)buf, len, flags
	};

	return (syscall(SYS_socketcall, SYS_RECV, args));
}

int 
native_recvfrom(int s, void *buf, size_t len, int flags,
		struct sockaddr *from, socklen_t *fromlen)
{
	unsigned long args[] = {
		s, (unsigned long)buf, len, flags,
		(unsigned long)from, (unsigned long)fromlen
	};

	return (syscall(SYS_socketcall, SYS_RECVFROM, args));
}

int 
native_recvmsg(int s, struct msghdr *msg, int flags)
{
	unsigned long args[] = {
		s, (unsigned long)msg, flags
	};

	return (syscall(SYS_socketcall, SYS_RECVMSG, args));
}

int
native_shutdown(int s, int how)
{
	unsigned long args[] = {
		s, how
	};

	return (syscall(SYS_socketcall, SYS_SHUTDOWN, args));
}

int 
native_fcntl(int fd, int cmd, long arg)
{
	return (syscall(SYS_fcntl, fd, cmd, arg));
}

pid_t
native_fork(void)
{
	return (syscall(SYS_fork));
}

int 
native_dup2(int old_fd, int new_fd)
{
	return (syscall(SYS_dup2, old_fd, new_fd));
}

int 
native_dup(int old_fd)
{
	return (syscall(SYS_dup, old_fd));
}

off_t 
native_lseek(int fd, off_t offset, int whence)
{
	return (syscall(SYS_lseek, fd, offset, whence));
}

ssize_t 
native_sendfile(int out_fd, int in_fd, off_t *offset, size_t count)
{
	return (syscall(SYS_sendfile, out_fd, in_fd, offset, count));
}
