/*   This is part of ALE4Net
 *   Application Level Environment for Networking
 *   
 *   Copyright 2004 Renzo Davoli University of Bologna - Italy
 *   
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License along
 *   with this program; if not, write to the Free Software Foundation, Inc.,
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */   
/*
 * Neelkanth Natu (neelnatu@yahoo.com) is the author of this file.
 * As long as you keep this blurb in the file, feel free to do whatever.
 */ 
#ifndef REQ_RESP_H
#define REQ_RESP_H

#if __WORDSIZE == 64
#error "This code assumes int and long are 32 bits"
#endif	/* if __WORDSIZE == 64 */

enum request_type {
	REQ_NONE = 0,
	REQ_SOCKET,
	REQ_CLOSE,
	REQ_BIND,
	REQ_LISTEN,
	REQ_ACCEPT,
	REQ_CONNECT,
	REQ_READ,
	REQ_WRITE,
	REQ_GETSOCKNAME,
	REQ_GETPEERNAME,
	REQ_RECV,
	REQ_RECVFROM,
	REQ_RECVMSG,
	REQ_SEND,
	REQ_SENDTO,
	REQ_SENDMSG,
	REQ_SHUTDOWN,
	REQ_SETSOCKOPT,
	REQ_GETSOCKOPT,
	REQ_FCNTL,
	REQ_SELECT,
	REQ_SELECT_STOP,
	REQ_FORK,
	REQ_DUP,

	REQ_IOCTL,
	REQ_LAST = REQ_IOCTL
};
extern const char *request_type_to_string[1 + REQ_LAST];

enum response_type {
	RESP_NONE = 0,
	RESP_SOCKET,
	RESP_CLOSE,
	RESP_BIND,
	RESP_LISTEN,
	RESP_ACCEPT,
	RESP_CONNECT,
	RESP_READ,
	RESP_WRITE,
	RESP_GETSOCKNAME,
	RESP_GETPEERNAME,
	RESP_RECV,
	RESP_RECVFROM,
	RESP_RECVMSG,
	RESP_SEND,
	RESP_SENDTO,
	RESP_SENDMSG,
	RESP_SHUTDOWN,
	RESP_SETSOCKOPT,
	RESP_GETSOCKOPT,
	RESP_FCNTL,
	RESP_SELECT,	/* For both REQ_SELECT and REQ_SELECT_STOP */
	RESP_FORK,
	RESP_DUP,

	RESP_IOCTL,
	RESP_LAST = RESP_IOCTL
};
extern const char *response_type_to_string[1 + RESP_LAST];

/* socket() request and response */
struct socket_req {
	int32_t req_type;
	int32_t domain;
	int32_t type;
	int32_t protocol;
};

struct socket_resp {
	int32_t resp_type;
	int32_t retval;
	int32_t alpine_errno;
};

/* close() request and response */
struct close_req {
	int32_t req_type;
	int32_t fd;
};

struct close_resp {
	int32_t resp_type;
	int32_t retval;
	int32_t alpine_errno;
};

/* bind() request and response */
struct bind_req {
	int32_t req_type;
	int32_t fd;		/* int sockfd */
	int32_t addrlen;	/* socklen_t addrlen */
	char myaddr[0];		/* struct sockaddr *myaddr */
};

struct bind_resp {
	int32_t resp_type;
	int32_t retval;
	int32_t alpine_errno;
};

/* listen() request and response */
struct listen_req {
	int32_t req_type;
	int32_t fd;
	int32_t backlog;
};

struct listen_resp {
	int32_t resp_type;
	int32_t retval;
	int32_t alpine_errno;
};

/* accept() request and response */
struct accept_req {
	int32_t req_type;
	int32_t fd;
	int32_t addrlen;
	char addr[0];
};

struct accept_resp {
	int32_t resp_type;
	int32_t addrlen;
	int32_t retval;
	int32_t alpine_errno;
	char addr[0];
};

/* connect() request and response */
struct connect_req {
	int32_t req_type;
	int32_t fd;
	int32_t addrlen;
	char addr[0];
};

struct connect_resp {
	int32_t resp_type;
	int32_t retval;
	int32_t alpine_errno;
};

/* ioctl() request and response */
struct ioctl_req {
	int32_t req_type;
	int32_t fd;
	int32_t command;
	int32_t data_length;
	char data[0];	/* optional data of size 'data_length' */
};

struct ioctl_resp {
	int32_t resp_type;
	int32_t retval;
	int32_t alpine_errno;
	char data[0]; 	/* optional data of size 'ioctl_req.data_length' */
};

/* read() request and response */
struct read_req {
	int32_t req_type;
	int32_t fd;
	int32_t count;
};

struct read_resp {
	int32_t resp_type;
	int32_t retval;
	int32_t alpine_errno;
	char data[0];	/* Present only if retval > 0, and of size 'retval' */
};

/* write() request and response */
struct write_req {
	int32_t req_type;
	int32_t fd;
	int32_t count;
	char data[0];	/* Present only if count > 0, and of size 'count' */
};

struct write_resp {
	int32_t resp_type;
	int32_t retval;
	int32_t alpine_errno;
};

/* getsockname/getpeername request and response */
struct getname_req {
	int32_t req_type;
	int32_t fd;
	int32_t namelen;
};

struct getname_resp {
	int32_t resp_type;
	int32_t namelen;
	int32_t retval;
	int32_t alpine_errno;
	char name[0];	/* Present only if retval == 0 and namelen > 0 */
};

/* recv request and response */
struct recvfrom_req {
	int32_t req_type;
	int32_t fd;
	int32_t len;
	int32_t flags;
	int32_t fromlen;
};

struct recvfrom_resp {
	int32_t resp_type;
	int32_t fromlen;
	int32_t retval;
	int32_t alpine_errno;
	char buf[0];	/* Optional and of length 'retval' */
	char from[0];	/* Optional and of length 'fromlen' */
};

/* recvmsg() request and response */
struct recvmsg_req {
	int32_t req_type;
	int32_t fd;
	int32_t namelen;
	int32_t controllen;
	int32_t buflen;	/* cumulative of the iovlens[] */
	int32_t flags;
};

struct recvmsg_resp {
	int32_t resp_type;
	int32_t namelen;
	int32_t controllen;
	int32_t flags;
	int32_t retval;
	int32_t alpine_errno;
	char name[0];	/* optional and of length 'msg_namelen' */
	char control[0];	/* optional and of length 'msg_controllen' */
	char buf[0];	/* optional and of length 'retval' */
};

/* send/sendto request and response */
struct sendto_req {
	int32_t req_type;
	int32_t fd;
	int32_t len;
	int32_t flags;
	int32_t tolen;
	char buf[0];	/* 'len' bytes long */
	char to[0];	/* 'tolen' bytes long */
};

struct sendto_resp {
	int32_t resp_type;
	int32_t retval;
	int32_t alpine_errno;
};

/* sendmsg() request and response */
struct sendmsg_req {
	int32_t req_type;
	int32_t fd;
	int32_t namelen;
	int32_t controllen;
	int32_t buflen;
	int32_t flags;
	char name[0];	/* optional and of length 'namelen' */
	char control[0];	/* optional and of length 'controllen' */
	char buf[0];	/* optional and of length 'buflen' */
};

struct sendmsg_resp {
	int32_t resp_type;
	int32_t retval;
	int32_t alpine_errno;
};

/* shutdown request and response */
struct shutdown_req {
	int32_t req_type;
	int32_t fd;
	int32_t how;
};

struct shutdown_resp {
	int32_t resp_type;
	int32_t retval;
	int32_t alpine_errno;
};

/* setsockopt() request and response */
struct setsockopt_req {
	int32_t req_type;
	int32_t fd;
	int32_t level;
	int32_t optname;
	int32_t optlen;
	char optval[0];		/* Optional and of length 'optlen' */
};

struct setsockopt_resp {
	int32_t resp_type;
	int32_t retval;
	int32_t alpine_errno;
};

/* getsockopt() request and response */
struct getsockopt_req {
	int32_t req_type;
	int32_t fd;
	int32_t level;
	int32_t optname;
	int32_t optlen;
	char optval[0];		/* Optional and of length 'optlen' */
};

struct getsockopt_resp {
	int32_t resp_type;
	int32_t optlen;
	int32_t retval;
	int32_t alpine_errno;
	char optval[0];		/* Optional and of length 'optlen' */
};

/* fcntl request and response */
struct fcntl_req {
	int32_t req_type;
	int32_t fd;
	int32_t cmd;
	int32_t arg;
};

struct fcntl_resp {
	int32_t resp_type;
	int32_t retval;
	int32_t alpine_errno;
};

/* 
 * Note: We don't specify struct timeval timeout because we don't depend on
 * the FreeBSD stack to do timeout for us. Rather we use the native Linux
 * select() for the timeout and simply call select_stop if the timeout
 * happens.
 */
struct select_req {
	int32_t req_type;
	int32_t nfd;
	fd_mask rdset[0];	/* size == howmany(nfd, sizeof(fd_mask)*8) */
	fd_mask wrset[0];
	fd_mask exset[0];
};

struct select_stop_req {
	int32_t req_type;
};

struct select_resp {
	int32_t resp_type;
	int32_t retval;
	int32_t alpine_errno;
	fd_mask rdset[0];
	fd_mask wrset[0];
	fd_mask exset[0];
};

/* fork() request and response messages */
struct fork_req {
	int32_t req_type;
	int32_t parent_pid; /* pid used by the server to identify the parent */
	int32_t child_pid;  /* pid used by the server to identify the child */
};

struct fork_resp {
	int32_t resp_type;
	int32_t retval;
	int32_t alpine_errno;
};

/* dup()/dup2() request and response messages */
struct dup_req {
	int32_t req_type;	/* REQ_DUP or REQ_DUP2 */
	int32_t old_fd;
	int32_t new_fd;		/* Set to -1 for REQ_DUP */
};

struct dup_resp {
	int32_t resp_type;	/* RESP_DUP or RESP_DUP2 */
	int32_t retval;
	int32_t alpine_errno;
};

extern enum request_type read_req_type(int fd);
extern enum response_type read_resp_type(int fd);
extern int write_resp_type(int fd, enum response_type resp_type);
extern int write_req_type(int fd, enum request_type req_type);
extern int read_n_bytes(int fd, char *buf, int n);
extern int write_n_bytes(int fd, const char *buf, int n);

#endif	/* ifndef REQ_RESP_H */
